package agents;


import environment.Cell;
import event.Event;
import event.RibosomeEvent;

import objects.FreedAgents;
import objects.MessengerRNA;
import utils.Constants;
import utils.SequenceUtils;

public class Ribosome implements Agent {

    
       public static int tmp=-1;
       private int id;
	private int boundTOmRNA;
	//private int mRNAposStart;
	//private int mRNAposEnd;
        private static boolean isFreeRibo=true;
	private int mRNAposA;
	private int noOfTranslations=1;
	private byte state;
	private int tRNAIDsiteA;
	private int tRNAIDsiteP;
	private boolean isCognate;
	private double lastReactionTime;
	private byte codonID; 
        private double startTime = -1; 	
	private double lastTranslocationTime;
	
	public int nearCognateAttempts;
	public int cognateAttempts;
	
	public int collisionsBeforeUnbinds;
	
	private int blockedBy;
	private int blocks;
	private int blocksIBS;
	/**
	 * empty contructor adds a free ribosome
	 * @param id
	 */
	public Ribosome(int id){
		this.id =id;
		boundTOmRNA = Constants.NONE;
		//mRNAposStart = -1;
		//mRNAposEnd = -1;
		mRNAposA = Constants.NONE;
		noOfTranslations=1;
		state = Constants.NONE;
		tRNAIDsiteA = Constants.NONE;
		tRNAIDsiteP = Constants.NONE;
		isCognate=true;
		lastReactionTime = 0;
		codonID = 0;
		lastTranslocationTime = 0;
		nearCognateAttempts = 0;
		cognateAttempts = 0;
		blockedBy =Constants.NONE;
		blocks = Constants.NONE;
		blocksIBS = Constants.NONE;
		collisionsBeforeUnbinds = Constants.NONE;
	}
	
	public int getState(){
		return this.state;
	}
	
	public boolean isCognate(){
		return isCognate;
	}
	
	
	/**
	 * return true if the ribosome is at the termination site and it transcribed enough times
	 * @param mRNA the strand that is attached to;
	 * @return
	 */
	public boolean canRelease(boolean atTerminationSite){
		return (atTerminationSite && noOfTranslations==0);
	}
	
	
	/**
	 * checks whether the ribosome can bind to the mRNA specieifed as parameter
	 * @param mRNA the mRNA where the ribosome wants to bind
	 * @return true if it can bind or false othwerwise 
	 */
	public boolean canBind(MessengerRNA mRNA){
		if (mRNAposA == Constants.NONE){
			if(mRNA.isOccupied(Constants.IBS_POSITION)==Constants.NONE){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * binds a ribosome to the mRNA with ID copyID
	 * @param mRNA the mRNA where to bind
	 * @param copyID, the ID of the mRNA where to bind
	 * @return true if binding succesful 
	 */
	public boolean bind(MessengerRNA mRNA, int copyID, int noOfTranslations){
		boolean result=false;
		int nextPos;
		 if (canBind(mRNA)){
			nextPos = 0;
			boundTOmRNA=copyID;
			this.noOfTranslations=noOfTranslations;
			//mRNA.setOccupancyState(nextPos, true);
			mRNAposA=nextPos;
			state = Constants.RIBOSOME_PERFORMING_RANDOM_WALK;
			tRNAIDsiteA = -1;
			tRNAIDsiteP = -1;
			result = true;
		}
		 
                  
		return result;
	}
	

	/**
	 * method that unbinds a ribosome from an mRNA. 
	 * The unbinding hapens only if the ribosome is bound to an mRNA and if it is bound to the IBS or termination site  
	 * @param mRNA the mRNA from where to bind
	 * @return true if unbound succeful or false otherwise
	 * 
	 */
	public boolean unbind(MessengerRNA mRNA){
		boolean result=false;
                //apparently Radu does not allow unbinding from the middle 
		//if(boundTOmRNA>=0 && (mRNAposA ==0|| mRNAposA==mRNA.getLength()-1)){
		if(boundTOmRNA>=0){
                boundTOmRNA=Constants.NONE;;
			noOfTranslations = Constants.NONE;;
			mRNAposA=Constants.NONE;;
			state = Constants.NONE;;
			lastTranslocationTime = Constants.NONE;;
			result = true;
		}
		 
		return result;
	}

	public int getElongationCurrentState(){
		return state;
	}
	
	/**
	 * returns the next translocation position
	 * @param length the length of the mRNA
	 * @return
	 */
	public int getNextPosition(int length){
		int nextPos = 0;
		
		if(boundTOmRNA>=0 && mRNAposA >=0){
			if(mRNAposA <= length-2){
				nextPos=mRNAposA+1;
			} else if(noOfTranslations > 0){
				nextPos=1;
			} else{
				nextPos=Constants.NONE;
			}
		} 
		
		return nextPos;
	}
	
	/**
	 * sets which ribosome is blocked by the current one
	 * @param ribosomeID
	 */
	public void setBlocks(int ribosomeID){
		blocks =ribosomeID; 
	}
	
	/**
	 * sets which ribosome is blocked by the current one residing at IBS
	 * @param ribosomeID
	 */
	public void setBlocksIBS(int ribosomeID){
		blocksIBS =ribosomeID; 
	}
	
	/**
	 * sets which ribosome blocks the current one
	 * @param ribosomeID
	 */
	public void setBlockedBy(int ribosomeID){
		blockedBy =ribosomeID; 
	}
	
	
	/**
	 * gets which ribosome is blocked by the current one
	 * @param ribosomeID
	 */
	public int getBlocks(){
		return blocks; 
	}
	
	/**
	 * gets which ribosome is blocked by the current one at IBS
	 * @param ribosomeID
	 */
	public int getBlocksIBS(){
		return blocksIBS; 
	}
	
	/**
	 * gets which ribosome blocks the current one
	 * @param ribosomeID
	 */
	public int getBlockedBy(){
		return blockedBy; 
	}
	
	
	/**
	 * checks whether a ribosome can slide on a mRNA and if yes it returns the next position.
	 * @param mRNA the mRNA on which to slide on
	 * @return the next position or -1 if there is no next position
	 */
	public int canTranslocate(MessengerRNA mRNA, int noOfSitesAfter, int noOfSitesBefore, boolean stallEventsWait){
		int after,nextPos = Constants.NONE;
		int blockingRibosome;
		blockedBy =Constants.NONE;
		if(mRNAposA>=0 && boundTOmRNA>=0){
			int length = mRNA.getLength();
			if(mRNAposA <= length-2){
				nextPos=mRNAposA+1;
				after = nextPos+noOfSitesAfter+noOfSitesBefore;
				if(after>length-1){after=length-1;}
			 
				if(mRNAposA <= after-nextPos+2){
					blockingRibosome = mRNA.areOccupied(nextPos,after);
					if(blockingRibosome != Constants.NONE){
						nextPos=mRNAposA;
						if(stallEventsWait){
							blockedBy = blockingRibosome;
						}
					}
				} else if(mRNAposA > after-nextPos+1){
					blockingRibosome =mRNA.isOccupied(after);
					if(blockingRibosome != Constants.NONE){
						nextPos=mRNAposA;
						if(stallEventsWait){
							blockedBy = blockingRibosome;
						}
					}
				}
			} else if(noOfTranslations > 0){
				nextPos=1;
				after = nextPos+noOfSitesAfter+noOfSitesBefore;
				if(after>length-1){after=length-1;}
				blockingRibosome =mRNA.areOccupied(nextPos,after);
				if(blockingRibosome != Constants.NONE){
					nextPos=mRNAposA;
					if(stallEventsWait){
						blockedBy = blockingRibosome;
					}
				}
				 
			}
			
		}
		
		
		 
		
		return nextPos;
	}
	
	
	/**
	 * slides a ribosome on the mRNA if posible. 
	 * Otherwise the ribosome unbinds from position IBS or termination site or doesn't do anything else from all the other positions 
	 * @param mRNA the mRNA on which the ribosome attempts to slide
	 * @return true if a slide was performed or false otherwise.
	 */
	public boolean translocate(MessengerRNA mRNA, int noOfSitesAfter, int noOfSitesBefore, boolean stallEventsWait, int nextPos){
		boolean result=false;
		//int nextPos = canTranslocate(mRNA,noOfSitesAfter,noOfSitesBefore, stallEventsWait);
		if(nextPos==mRNAposA || nextPos==Constants.NONE){
			//if(mRNAposA==0 || mRNAposA ==  mRNA.getLength()-1 ){
			//	unbind(mRNA);
			//}
			result = false;
		} else{
			//mRNA.setOccupancyState(mRNAposA, false);
			//mRNA.setOccupancyState(nextPos, true);
			mRNAposA = nextPos;
			result = true;
			//this.tRNAID = -1;
			this.state=Constants.RIBOSOME_ELONGATION_STATE_EMPTY;
		}
	



	return result;
	}
	
	
	
	/**
	 * generates a string with the ribosome state
	 * @return a string with the sate of the ribosome 
	 */
	public String toString(){
		if(boundTOmRNA==-1){
			return "Ribosome: "+id+", "+ " free ";
		}
		
		String stateStr="";
		if(state==Constants.RIBOSOME_ELONGATION_STATE_EMPTY){
			stateStr = " with no tRNA bound";
		} else if(state==Constants.RIBOSOME_ELONGATION_STATE_BOUND){
			stateStr = " with a tRNA of type t"+this.tRNAIDsiteA+" bound";
		} else if(state==Constants.RIBOSOME_ELONGATION_STATE_TRANSFER){
			stateStr = " with a tRNA of type t"+this.tRNAIDsiteA+" bound and the amino-acid transfered";
		}

		String strBlocks="";
		if(blocks!=Constants.NONE){
			strBlocks+=", blocks "+blocks;
		}
		if(blocksIBS!=Constants.NONE){
			strBlocks+=", blocks at IBS "+blocksIBS;
		}
		if(blockedBy!=Constants.NONE){
			strBlocks+=", is blocked by "+blockedBy;
		}
		return "Ribosome: "+id+", "+ "bound to mRNA: " + boundTOmRNA +", at position:"+ mRNAposA + " codonId: "+codonID+" "+stateStr+strBlocks;
	}
	
        
        public static boolean existFreeRibos(){return isFreeRibo;};
        
        public static void  setIsFreeRibos(boolean what){ isFreeRibo=what;};
        
        
	/**
	 * returns the mRNA on which the ribosomes is bound
	 * @return
	 */
	public int getmRNA(){
		return boundTOmRNA;
	}
	
	/**
	 * returns the position on the mRNA of the ribosome
	 * @return
	 */
	public int getmRNAPosition(){
		return mRNAposA;
	}
	
	/**
	 * removes one translation and returns the number of translations left;
	 * @return
	 */
	public int updateTranslated(){
		noOfTranslations--;
		return noOfTranslations;
	}
	
	/**
	 * returns the number of translations left to complete.
	 * @return
	 */
	public int getNoOfTranslations(){
		return noOfTranslations;
	}
	
	/**
	 * returns true if the ribosome is bound to an mRNA
	 * @return
	 */
	public boolean isBound(){
		return this.boundTOmRNA>=0 && this.mRNAposA>=0;
	}
	
	/**
	 * returns the time of the last reaction
	 * @return
	 */
	public double getLastReactionTime(){
		return lastReactionTime;
	}

	/**
	 * returns the id of the bound tRNA
	 * @return
	 */
	public int gettRNAIDsiteA(){
		return this.tRNAIDsiteA;
	}
	
	/**
	 * returns the id of the bound tRNA
	 * @return
	 */
	public int gettRNAIDsiteP(){
		return this.tRNAIDsiteP;
	}
	
	/**
	 * the ribosome is an agent and it acts
	 * @param cp the reference to the cell
	 * @return an object FreedAgents which identifies whether something got freed
	 */
	public FreedAgents act(RibosomeEvent e, Cell cp){
		FreedAgents gotFree = new FreedAgents();
		
              //if(tmp==-1) tmp=id;
         // if(id==tmp)
 //if(mRNAposA >1)               
  //  System.out.println(cp.getCellTime() + "  " + id + "  " + mRNAposA  + "   " + e.mRNAID + " " + state + " e:state " + e.state + " e.trnaid  " + e.tRNAID  + " Ribos Waiting:  "+ (cp.getEventList()).gettRNABindingQ().getNumberOfWaitingRibos() );
              
 
   
                    
                if(this.boundTOmRNA<0){
      
	  
                    gotFree = performRibosomeBinding(e,cp);
                      
    
     
     
                        
		} else if(this.mRNAposA==Constants.IBS_POSITION){
                    
               
			gotFree=performRandomWalk(e,cp);
                      
                        
		} else{
			// the ribosome is bound so it will attempt to elongate
		 
                    
                    performElongation(e,cp);
                        
                        //schedule the next event
                         
                      //  cp.scheduleNextRibosomeEvent(e.ribosomeID, e.time, gotFree);
		}
		//update  cell time
		//cp.updateCellTime(e.time);
		
		//update the time of the last reaction of current ribosome
		this.lastReactionTime = e.time;
		
	
		
		//if(this.boundTOmRNA>=0){
		//	cp.scheduleNextRibosomeEvent(e.ribosomeID, e.time, gotFree);
		//}
		
		return gotFree;
	}
	
	
	/**
	 * returns the current codon
	 * @return
	 */
	public byte getCodon(){
		return codonID;
	}
	
	/**
	 * perform a ribosome mRNA binding
	 * it precomputes the number of translations once bound as either a maximum value if (unbindsFromterminationIfFree) or a random number if the ribosome cannot unbind from the termination site
	 * @param e the ribosome binding event that needs to be performed
	 */
	private FreedAgents performRibosomeBinding(RibosomeEvent e, Cell cp){
		
             
		FreedAgents gotFree = new FreedAgents();
		
		int noOfTranslations = cp.getNoOfTranslationsPerRibosomemRNAMax();
		
		  
		if(!cp.unbindsFromterminationIfFree()){
			noOfTranslations = cp.getNoOfTranslationsPerRibosomemRNAMin() + cp.getRandomGenerator().nextInt(cp.getNoOfTranslationsPerRibosomemRNAMax() - cp.getNoOfTranslationsPerRibosomemRNAMin());
		}
		
		//change the ribosome params
		boolean bound = bind(cp.getmRNA(e.mRNAID), e.mRNAID, noOfTranslations);
		//mark the mRNA IBS site as occupied
		if(bound){  
				//FreedAgents buffer = new FreedAgents();
				
				if(cp.occupyRibosome(e.ribosomeID)){ 
					gotFree.ribosome = e.ribosomeID;
				} 
                                
                                
                                
				if(cp.occupymRNA(e.mRNAID)){ 
					gotFree.mRNA = e.mRNAID;
					cp.occupymRNAPosition(e.mRNAID, getmRNAPosition(), e.ribosomeID);
				}

				 
				//updatePropensities(buffer);
				 if(cp.printsPolyosomeSize()){
					cp.incrementPolyosomeSize(e.time);
				}
				
				cp.incrementRibosomesPerMRNA(e.mRNAID,e.time);


//also schedule a random walk event
  (cp.getEventList()).scheduleRandomWalk(e.ribosomeID, cp.getCellTime(),    cp.getmIBSMoveRate(e.mRNAID),cp.getRandomGenerator(), e.mRNAID, e.lastReactionTime, cp);
  
               

		}
           
//now schedule a new event ribosome binding to tRNA , no need to update everything because only the number of 
 //of ribosomes has changed. 
if(Ribosome.existFreeRibos()){
           (cp.getEventList()).scheduleRibosomeBinding(cp,cp.getmRNAType(e.mRNAID));      


}
 	
		return gotFree;
	}

        
        
        
        
        //*************************
        //*************************
        //*************************
        //This function binds tRNA, either Cog, or near Cog. 
        //It then also schedules a new event
        //*************************
        //*************************
        //*************************
        //*************************
        
        
        
        private void bindtRNA(RibosomeEvent e, Cell cp){
		
         
                
			this.state=Constants.RIBOSOME_ELONGATION_STATE_BOUND;
			this.tRNAIDsiteA = e.tRNAID;
			cp.settRNABusy(e.tRNAID);
			 
                        cp.recordTRNAProportionsAvg(e.tRNAID, e.time);
			//cp.decrementNumberOfFreeBoundRibosomes();
			 
			
                        this.isCognate = SequenceUtils.tRNAtable.isCognate(this.codonID, this.tRNAIDsiteA);
			
                        if(!this.isCognate){
				this.nearCognateAttempts++;
			} else{
				this.cognateAttempts++;
			}
//I suspect we do not need this			//cp.updateElongationEventQueueBindingCodon(e.tRNAID,e.time);
 
                        
 //having BOUND the tRNA, now schedule the UNBINDING
(cp.getEventList()).scheduleInternalRNATransfer(e.ribosomeID, cp.getCellTime(), e.mRNAID, this.mRNAposA, e.tRNAID, e.codonID,  isCognate, cp.getRandomGenerator(), lastReactionTime, cp);
                                 
 

//re-compute the binding affinities:  
   ((cp.getEventList()).getRibo2MRNAEventQ()).updateRibosomeBindingPropensities(cp.getmRNAType(e.mRNAID), cp);
   
   
   //add another event of tRNA to ribosome binding
   ((cp.getEventList()). gettRNABindingQ()).addNexttRNAToRiboBindingEvent(cp.getCellTime(),cp.getRandomGenerator(), cp);
  
  
           
                
        }
        
        
        
        
        
          //*************************
        //*************************
        //*************************
        //This function releases the tRNA that has previously been bound
        //*************************
        //*************************
        //*************************
        //*************************
        
        
        
        private void releasetRNA(RibosomeEvent e, Cell cp){
		 
           
                
                
		    
                    this.state=Constants.RIBOSOME_ELONGATION_STATE_EMPTY;
			cp.resettRNABusy(this.tRNAIDsiteA);
//			cp.incrementNumberOfFreeBoundRibosomes();
                        //cp.decrementNumberOfFreeBoundRibosomes();
                        //This next line recomputes the  propensities for the tRNA binding (because tRNA numbers have just changed)
			//cp.updateElongationEventQueueBindingCodon(this.tRNAIDsiteA,e.time);
                 
//we must now schedule a new tRNA binding event                        
((cp.getEventList()). gettRNABindingQ()).scheduleNextElongationBindingEvent(e.ribosomeID,cp.getCellTime(), e.mRNAID,this.getmRNAPosition(), e.codonID,cp.getRandomGenerator(),cp);
         		this.tRNAIDsiteA = -1;
                        
                        
                        
                        
                   
                
        }
        
        
        
               
          //*************************
        //*************************
        //*************************
        //This function readies the ribosome to move on to next position
        // NOte it does not actually move on....that requires an extra step
        //*************************
        //*************************
        //*************************
        //*************************
        
        
        
        private void ribosomeFinishReading(RibosomeEvent e, Cell cp){
		
            boolean tmp=false;
            
             
                    this.state=Constants.RIBOSOME_ELONGATION_STATE_TRANSFER;
			//cp.unchargetRNA(this.tRNAIDsiteA); This should not be here. I moved it into moveOn()
			cp.recordTRNAProportionsAvg(e.tRNAID, e.time);
			//this.tRNAID = -1;
			if(this.isCognate){
				cp.incrementCognateElongated(this.codonID,e.time);
			} else{
				cp.incrementNearCognateElongated(this.codonID,e.time);
			}
		
                //So, here now we need to schedule a new event to translocate
              
 	    
  
                        //skip scheduling a separate step, if parameters are set in that way
 if(cp.getElongationTranslocationRate()  <0) {
 
     RibosomeEvent riboEv = (cp.getEventList()).getTranslocationQ().getMoveOnEvent( e,  cp);
     
     
     if(riboEv.codonID!=-1) tmp=ribosomeMoveOn(riboEv,cp);
    
    //we make an exception when the Ribosome is in pos zero. This is to avoid unnecessary ping-pong. 
    if(riboEv.codonID==-1)  {
        
        (cp.getEventList()).getTranslocationQ().scheduleNextMovement(e.ribosomeID,cp.getCellTime(), cp.getElongationTranslocationRate(), e.mRNAID, lastReactionTime, cp, cp.getRandomGenerator());        

       
    }
 
 
 }           
 else{                                      
(cp.getEventList()).getTranslocationQ().scheduleNextMovement(e.ribosomeID,cp.getCellTime(), cp.getElongationTranslocationRate(), e.mRNAID, lastReactionTime, cp, cp.getRandomGenerator());        
         
}          
                
                        
                         
        }
        
        
        
        
        
        
        
            //*************************
        //*************************
        //*************************
        //This function actually moves on the ribosome
        //*************************
        //*************************
        //*************************
        //*************************
        
        
        
        private boolean ribosomeMoveOn(RibosomeEvent e, Cell cp){
		 boolean success;
		   
                   
               
                     		int originalPos = this.mRNAposA;
			double lLastTranslocationTime = this.lastTranslocationTime;
			boolean lwasCognate = this.isCognate;
			int lcodonID = this.codonID;
		 
                    
                        success=performTranslocation(e,cp);
		
                        // this is inconsequential rubbish...delete soon(maybe Chris can delete this).
                        if(this.mRNAposA!=originalPos){
				
                            if(lwasCognate){
					cp.addCognateElongationTimeTotal(lcodonID, e.time-lLastTranslocationTime,e.time);
				} else{
					cp.addNearCognateElongationTimeTotal(lcodonID, e.time-lLastTranslocationTime,e.time);
				}
			
                            
                          //  cp.incrementNumberOfFreeBoundRibosomes();
				
                            
                            
				// release the P site tRNA and schedule it for recharge and then move the tRNA from site A to site P
				if(this.tRNAIDsiteP!=Constants.NONE){
                                        cp.unchargetRNA(this.tRNAIDsiteP); 
					cp.scheduleNexttRNAEvent(this.tRNAIDsiteP, e.time);
                                       this.tRNAIDsiteP=Constants.NONE; 
				} 
                                
                                

    if(this.boundTOmRNA!=Constants.NONE){
            this.tRNAIDsiteP = this.tRNAIDsiteA;
            this.tRNAIDsiteA = Constants.NONE;
                                        }  
    else{

            if(this.tRNAIDsiteP != Constants.NONE)
            cp.scheduleNexttRNAEvent(this.tRNAIDsiteP, e.time);

            if(this.tRNAIDsiteA != Constants.NONE)
             cp.resettRNABusy(tRNAIDsiteA);           
            this.tRNAIDsiteA = Constants.NONE;
              this.tRNAIDsiteP = Constants.NONE;                             
           
                                                }

				
				 
				cp.addCognateAttemptsTotal(lcodonID, this.cognateAttempts,e.time);
				cp.addNearCognateAttemptsTotal(lcodonID, this.nearCognateAttempts,e.time);
				this.nearCognateAttempts=0;
				this.cognateAttempts = 0;
                                
                        
                              
                                
                                
                                return true;   
			}
                   
                      
            return false;
                   
            
                
        }
        
        
        
        
        
        
        
        
        
        //****************************************
        
        
        //*************************
        //*************************
        //*************************
        //*************************
        //*************************
        //*************************
        
	
	private void performElongation(RibosomeEvent e, Cell cp){
		 
		 
		//binds a cognate/near-cognate tRNA
		if(this.state == Constants.RIBOSOME_ELONGATION_STATE_EMPTY && e.state==Constants.RIBOSOME_ELONGATION_STATE_BOUND){
                    
                  //this binds the tRNA and schedules an unbinding event
                     bindtRNA(e,cp);
                    
                    
                             
                        
		}
		
		//cancels peptidyl transfer tRNA unbinds
		if(this.state == Constants.RIBOSOME_ELONGATION_STATE_BOUND && e.state==Constants.RIBOSOME_ELONGATION_STATE_EMPTY){
			
                     
              
       //This releases the tRNA and schedules a new binding event
                   releasetRNA(e,cp);
                        
                        
                        
                        
                        
		}
		
		//performs peptidyl transfer  tRNA remains and the 
		if(this.state == Constants.RIBOSOME_ELONGATION_STATE_BOUND && e.state==Constants.RIBOSOME_ELONGATION_STATE_TRANSFER){
		
                    
                 
                    ribosomeFinishReading(e,cp);
                
                
             
                
                
                }
		
		//translocates: We cannot quite forgte about this, even when the translocation rate is negative
                //when the ribo is stuck, then this will still be used. 
		if(this.state == Constants.RIBOSOME_ELONGATION_STATE_TRANSFER && e.state==Constants.RIBOSOME_ELONGATION_STATE_EMPTY){
	
                     
                    ribosomeMoveOn(e,cp);
                     
                    
                    
		}
		 
	}
	
        
        
        
        
        
        
        
        
	/**
	 * perform a translocation step
	 * if the ribosome is at termination site and the sliding is blocked then it is released 
	 * if the ribosome is at IBS and it slided then the mRNA gets free
	 * @param RibosomeEvent the events that is performed
	 * @return  a 2 dim array of int 1) -1 if ribosomes remains bound or the id of the ribosome if this gets unbound 2) the ID of the mRNA if the mRNA IBS site gets free or -1 otherwise
	 */
	private boolean performTranslocation(RibosomeEvent e, Cell cp){
		FreedAgents gotFree=new  FreedAgents();
//		the original position before sliding
		
	MessengerRNA tyu;
		int originalPosition = getmRNAPosition();
		
		int mRNAlength = cp.getmRNAlength(e.mRNAID);
		boolean originalAtTerminationSite = false;
		if(!originalAtTerminationSite && originalPosition ==  mRNAlength-1 ){
			originalAtTerminationSite = true;
		}
		// attempt to perform sliding
int cm = cp.canMove(e.ribosomeID);

 
boolean itTranslocated = translocate(cp.getmRNA(this.boundTOmRNA),cp.getNoOfRibosomeSitesAfter(),cp.getNoOfRibosomeSitesBefore(), cp.doesStallEventsWait(), cm);

		// get new position
                //Note that the internal variable for the next position has been changed by 
                //translocate, which is why nextPos and originalPosition are actuially different
int nextPosition = getmRNAPosition();
		
		 
//First large block: What happens if the translocation was unsuccessful
if(!itTranslocated){ 	

 //if it did not slide but it can release from initial or last position then release it
 collisionsBeforeUnbinds--;
                        
                        
boolean cond1=canRelease(originalAtTerminationSite);
boolean cond2= (originalAtTerminationSite && cp.unbindsFromterminationIfFree());
boolean cond3= !cond1 && !cond2 &&  collisionsBeforeUnbinds==0;
   
           
//If it goes into the next block, the ribo is released                        
if(cond1 || cond2 ||cond3 ){
				MessengerRNA tye = cp.getmRNA(this.boundTOmRNA);
				int mymRNA= boundTOmRNA;
				unblockRibosomes(cp, e.time,gotFree, e.ribosomeID);//if unbinds then release blocked ribosomes
				gotFree= cp.releaseRibosome(e.ribosomeID,e.time);
                                //if gotFree.ribosome > -1 then the ribo is still bound.
                                
				if(cp.printsPolyosomeSize()){
					cp.decrementPolyosomeSize(e.time);
				}
			
//CHU only print when you actually finished the translation	
if(cond1 || cond2){                                
System.out.format("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",tye.getName(),mRNAlength,startTime, e.time, e.time-startTime, (e.time-startTime)/mRNAlength,cp.reportNumberOfRibsomesOn(mymRNA),cp.getFractionOfFreeRibosomes()); 
}

startTime = -1;	


cp.removeRibosomeFrom(mymRNA);
//********** 
} ///end of release block
                else{    
       (cp.getEventList()).getTranslocationQ().scheduleNextMovement(e.ribosomeID,cp.getCellTime(), cp.getElongationTranslocationRate(), e.mRNAID, lastReactionTime, cp, cp.getRandomGenerator());        
        }



			cp.incrementUnSuccessfulElongation(e.time);
			
			//do not stay forever at the termination site... just unbind
			if(originalAtTerminationSite && !cp.unbindsFromterminationIfFree()){
				updateTranslated();
		                                                                          }
			cp.incrementUnsuccesfullTranslocationCount(e.time);
                                        
		} //Second block: What happens when translocation was successful
 else{
			  
			this.codonID=e.codonID;
                     
                        
			cp.slideRibosomeOnmRNA(getmRNA(), originalPosition, nextPosition);
			cp.incrementSuccessfulElongation(e.time);		
	
                        
                        
                        
                        
// if elongated from last position then mark a new translation as finished
			if(originalAtTerminationSite){
				updateTranslated();
				cp.incrementTranslated(e.mRNAID,e.time);
//CHU	
System.out.format("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",(cp.getmRNA(this.boundTOmRNA)).getName(), mRNAlength,startTime, e.time, e.time-startTime, (e.time-startTime)/mRNAlength,cp.reportNumberOfRibsomesOn(e.mRNAID),cp.getFractionOfFreeRibosomes()); 
startTime = e.time;	
//**********
                                                     }
                        
                        
                    
                     
                        
			this.lastTranslocationTime = e.time;	
			cp.incrementSuccesfullTranslocationCount(e.time);
			//if this ribosome is blocking another ribosome and it moved then check whether the other ribosome can be moved
			unblockRibosomes(cp, e.time, gotFree, e.ribosomeID);
			collisionsBeforeUnbinds = cp.getCollisionsBeforeUnbinds();
  
                        //schedule a new tRNA binding event if we have just moved on to a new site successfully                    
      
     (cp.getEventList()).gettRNABindingQ().scheduleNextElongationBindingEvent(e.ribosomeID,cp.getCellTime(), e.mRNAID, mRNAposA, e.codonID, cp.getRandomGenerator(), cp);
             
    
     if(!originalAtTerminationSite){  
     //sanity check
     if(getmRNAPosition()<0){System.out.println("\n Error in  &BFR%%DM<<<"); System.exit(-1);}
                                      }  
    
    
                        
		} //end second block
		
		return itTranslocated;
	}
	
	
	/**
	 * perform random walk on IBS
	 * if the ribosome slided then the mRNA gets free
	 * @param RibosomeEvent the events that is performed
	 * @return  a 2 dim array of int 1) -1 if ribosomes remains bound or the id of the ribosome if this gets unbound 2) the ID of the mRNA if the mRNA IBS site gets free or -1 otherwise
	 */
	private FreedAgents performRandomWalk(RibosomeEvent e, Cell cp){
		FreedAgents gotFree=new  FreedAgents();
		//the original position before sliding
		int originalPosition = getmRNAPosition();
boolean reScheduleRiboBinding=false;
		
		// attempt to perform sliding
		//byte codonID=cp.getCodon(this.boundTOmRNA, this.mRNAposA);
		//boolean isCognate=true;
		boolean itTranslocated = translocate(cp.getmRNA(this.boundTOmRNA),cp.getNoOfRibosomeSitesAfter(),cp.getNoOfRibosomeSitesBefore(), cp.doesStallEventsWait(), cp.canMove(e.ribosomeID));

		// get new position
		int nextPosition = getmRNAPosition();
		//if it slided chenage the occupancy markings on the mRNA strand
	
                 
		if(!itTranslocated){
			//if it did not slide but it can release from initial or last position then release it
			collisionsBeforeUnbinds--;
			if(cp.unbindsFromIBSifFree() || collisionsBeforeUnbinds==0){
				unblockRibosomes(cp, e.time,gotFree, e.ribosomeID);//if unbinds then release blocked ribosomes
				gotFree= cp.releaseRibosome(e.ribosomeID, e.time);
				if(cp.printsPolyosomeSize()){
					cp.decrementPolyosomeSize(e.time);
                                      
				}
                     
                                reScheduleRiboBinding=true ;     
                  
                                
			}else{//schedule another random walk if  the first one has failed. 
 (cp.getEventList()).scheduleRandomWalk(this.id ,cp.getCellTime(),cp.getmIBSMoveRate(this.boundTOmRNA), cp.getRandomGenerator(),this.boundTOmRNA, this.getLastReactionTime() ,  cp);
                           
                        }
                        
                        cp.incrementUnsuccesfullEventRandomWalkCount(e.time);
		} else{
                    
                    reScheduleRiboBinding=true;
			 this.codonID = cp.getCodon(this.boundTOmRNA, this.mRNAposA);
			//this.codonID = e.codonID;
			//it released the IBS site so mark this/
			if(originalPosition != nextPosition){
				cp.slideRibosomeOnmRNA(getmRNA(), originalPosition, nextPosition);
			//	cp.incrementNumberOfFreeBoundRibosomes();
			}
                         
                        
			cp.freemRNA(e.mRNAID);
			gotFree.mRNA = e.mRNAID;
			this.lastTranslocationTime = e.time;	
//CHU
startTime = e.time;
cp.addRibosomeTo(gotFree.mRNA);
//System.out.format("&&&&&&  %s\t%s \n", startTime, gotFree.mRNA); 
//--------------------

			cp.incrementSuccesfullEventRandomWalkCount(e.time);
			unblockRibosomes(cp, e.time, gotFree, e.ribosomeID);
			collisionsBeforeUnbinds = cp.getCollisionsBeforeUnbinds();
         
                        
               
                        
  //schedule the next event: binding of tRNA
 (cp.getEventList()).gettRNABindingQ().scheduleNextElongationBindingEvent(e.ribosomeID,cp.getCellTime(), e.mRNAID, mRNAposA, codonID, cp.getRandomGenerator(), cp);
   
 
    
                
                } 
		
                
        //now schedule a new event ribosome binding to tRNA , no need to update because only the number of 
 //of ribosomes has changed. 
if(Ribosome.existFreeRibos() && reScheduleRiboBinding){
           (cp.getEventList()).scheduleRibosomeBinding(cp,cp.getmRNAType(e.mRNAID));      


}        
                
                
		return gotFree;
	}
	
	/**
	 * attempts to unblock the ribosomes stuck behind
	 * @param cp
	 * @param time
	 * @param gotFree
	 */
	public void unblockRibosomes(Cell cp, double time,FreedAgents gotFree, int ribosomeID){
		//blocks ribosome
		if(blocks!=Constants.NONE){
			if(cp.isRibosomeFree(blocks)){
				//cp.printDebugInfo("ribosome "+blocks+" got free meanwhile");
				blocks=Constants.NONE;
			} else{
				
                            System.err.println("Ribosome.java unblock Ribosome: We should not be here");
                            System.exit(-1);
                            //cp.scheduleNextRibosomeEvent(blocks, time);
			}
		}
		
		//blocks IBS
		if(blocksIBS!=Constants.NONE){
			if(cp.isRibosomeFree(blocksIBS)){
				//cp.printDebugInfo("ribosome "+blocks+" got free meanwhile");
				blocksIBS=Constants.NONE;
			} else{
				//cp.printDebugInfo("rw: " + e.ribosomeID+" blocks "+blocks);
				//cp.printDebugInfo(time+": ribosome " + blocksIBS + " unblocked by ribosome" + ribosomeID);
                               System.err.println("Ribosome.java blockIBS Ribosome: We should not be here");
                            System.exit(-1);
				//cp.scheduleNextRibosomeEvent(blocksIBS, time);
			}
		}
	}
	
}
