package event;

import java.util.PriorityQueue;

import objects.FreedAgents;

import environment.Cell;

public class BindRibo2mRNAQ {
	
	protected double[] ribosomeBindingPropensity;// the propensity that ribosomes bind to various mRNA species
	protected double ribosomeBindingPropensitySum;

	private RibosomeEvent bindingEvent;

	/**
	 * class constructor
	 * @param cp
	 */
	public BindRibo2mRNAQ(Cell cp){
		//binding event
		this.bindingEvent = null;
		
	
		// update the propensities of the ribosome+mRNA interactions
		this.ribosomeBindingPropensitySum = 0;
		this.ribosomeBindingPropensity = new double[cp.getNoOfmRNATypes()];
		int freeRibosomes = cp.getNoOfFreeRibosomes();
		for(int i=0;i<cp.getNoOfmRNATypes();i++){
			this.ribosomeBindingPropensity[i]=cp.getmRNACopyNumber(i)*freeRibosomes*cp.getmRNARibosomeAffinity(i);
			this.ribosomeBindingPropensitySum+=this.ribosomeBindingPropensity[i];
		}
	}
	
	
	
	/**
	 * returns the next binding event
	 * @return
	 */
	public RibosomeEvent peek(){
		return bindingEvent;
	}
	
	/**
	 * returns the next binding event and removes it from the list
	 * @return
	 */
	public RibosomeEvent pop(){
		RibosomeEvent re=bindingEvent;
		bindingEvent = null;
		return re;
	}
	
	/**
	 * replaces the current binding event with a new one
	 * @param re the new event
	 */
	public void add(RibosomeEvent re){
		bindingEvent=re;
                 
	}
	
	
	/**
	 * returns true if the binding event is null
	 * @return
	 */
	public boolean isEmpty(){
		return bindingEvent==null;
	}
	
        
        
        
        /**
	 * updates the RibsomeBinding Propensities (simple and slow version); added by Chu
         *   * The propensities are calculated here without taking account
         * of free ribosomes. So the formula
         * cp.getFreemRNACopyNumber(mRNAtype)*cp.getmRNARibosomeAffinity(mRNAtype)
         * should be multiplied with the number of free ribos.
         * This will be done only before scheduling...to save time. 
	 * @param gotFree
	 */
	public void updateRibosomeBindingPropensities(Cell cp){
            
		int freeRibosomes=  cp.getNoOfFreeRibosomes();
		
		// recompute propensities
			 
			
			ribosomeBindingPropensitySum =0;
			for(int i=0;i<ribosomeBindingPropensity.length;i++){
				ribosomeBindingPropensity[i]=cp.getFreemRNACopyNumber(i)*cp.getmRNARibosomeAffinity(i);
				ribosomeBindingPropensitySum+=ribosomeBindingPropensity[i];
			}
			//ribosomeBindingPropensitySum = propensitySum();
		
	}
        
        
        
     
        
        
        
	/**
	 * updates the RibsomeBinding Propensities
         * NOTE!******************************
         * The propensities are calculated here without taking account
         * of free ribosomes. So the formula
         * cp.getFreemRNACopyNumber(mRNAtype)*cp.getmRNARibosomeAffinity(mRNAtype)
         * should be multiplied with the number of free ribos.
         * This will be done only before scheduling...to save time. 
         * *************************************
	 * @param gotFree
	 */
	public void updateRibosomeBindingPropensities(int mRNAtype, Cell cp){
		// if just the number of mRNAs changed then smart update of one propensity
		
            
            
		if(mRNAtype>0){
			//int mRNAtype = cp.getmRNAType(mRNAID);
			 
			//remove old propensity from propensity sum
			ribosomeBindingPropensitySum-=ribosomeBindingPropensity[mRNAtype];
			//compute new propensity
			ribosomeBindingPropensity[mRNAtype]=cp.getFreemRNACopyNumber(mRNAtype)*cp.getmRNARibosomeAffinity(mRNAtype);
			//add it to the sum
			ribosomeBindingPropensitySum+=ribosomeBindingPropensity[mRNAtype];
                        
              
   //System.out.println("--------------------> mrna copy num:" + cp.getFreemRNACopyNumber(mRNAtype)   + " affinity " + cp.getmRNARibosomeAffinity(mRNAtype));
                                  
                        
		} else{
		// recompute propensities
			
			
			ribosomeBindingPropensitySum =0;
			for(int i=0;i<ribosomeBindingPropensity.length;i++){
				ribosomeBindingPropensity[i]=cp.getFreemRNACopyNumber(i)*cp.getmRNARibosomeAffinity(i);
				ribosomeBindingPropensitySum+=ribosomeBindingPropensity[i];
			}
			//ribosomeBindingPropensitySum = propensitySum();
//System.out.println("--------------------> PropensitySum:" + ribosomeBindingPropensitySum  );
   		
                
                }
	}
	
	/**
	 * computes the sum of all propensities as in Gillespie algorithm
	 * @return
	 */
	private double propensitySum(){
		double result=0;
		for(double d:ribosomeBindingPropensity){
			result+=d;
		}
		return result;
	}
	
	/**
	 * returns the propensity sum
	 * @return
	 */
	public double getPropensitySum(){
		return this.propensitySum();
	}
	
	/**
	 * returns the array of all propensities
	 * @return
	 */
	public double[] getPropensities(){
		return this.ribosomeBindingPropensity;
	}
	
	public void clear(){
		this.bindingEvent = null;
	}
}

