package event;

import agents.Ribosome;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Random;

import objects.FreedAgents;
import objects.TRNAPropensity;

import environment.Cell;

import utils.Constants;
import utils.Gillespie;

public class EventList {
	//private PriorityQueue<RibosomeEvent> elongationEventQueue;
	
	private PriorityQueue<RibosomeEvent> elongationTransferEventQueue;
	private PriorityQueue<TRNAEvent> tRNAEventQueue;

	
	//private PriorityQueue<RibosomeEvent> elongationTranslocationEventQueue;
	//private PriorityQueue<RibosomeEvent> randomWalkEventQueue;

	private RibosomeMovementEventQueue elongationTranslocationEventQueue;
	private RibosomeMovementEventQueue randomWalkEventQueue;

	
	private BindRibo2mRNAQ bindRibo2mRNAQ; 
	private ElongationBindingEventQueueDM elongationBindingEventQueue;
	
	
	private boolean isGillespieDirect;
	
	private double tRNAbindingRate;
	private boolean nonGognateIntroduceWaitingTime;
	private double  nonGognateWaitingTime;
	private double  nearGognateWaitingTime;
	private boolean nonCognateChangeBindingRate;
	private double  nonCognateMMConstant;
	
	private double elongationCognateRate;
	private double elongationCognateProbability;
	private double elongationNearCognateRate;
	private double elongationNearCognateProbability;
	private double elongationNearCognateProbabilityExitRecognition;
	private double elongationCognateProbabilityExitRecognition;

	private double elongationTranslocationRate;
	

	
	//cell parameters
	
	public EventList(Cell cp){
		// cell params
		this.isGillespieDirect = cp.doesElongationBindingEventsUseGillespeDirect();
		this.tRNAbindingRate = cp.gettRNAbindingRate();
		this.nonGognateIntroduceWaitingTime = cp.doesNonGognateIntroduceWaitingTime();
		this.nonGognateWaitingTime = cp.getNonGognateWaitingTime();
		this.nearGognateWaitingTime = cp.getNearGognateWaitingTime();
		this.nonCognateChangeBindingRate = cp.doesNonCognateChangeBindingRate();
		this.nonCognateMMConstant = cp.getNonCognateMMConstant();
		
		this.elongationCognateRate = cp.getElongationCognateRate();
		this.elongationCognateProbability = cp.getElongationCognateProbability();
		this.elongationNearCognateRate = cp.getElongationNearCognateRate();
		this.elongationNearCognateProbability = cp.getElongationNearCognateProbability();
		this.elongationNearCognateProbabilityExitRecognition = cp.getElongationNearCognateProbabilityExitRecognition();
		this.elongationCognateProbabilityExitRecognition = cp.getElongationCognateProbabilityExitRecognition();

		this.elongationTranslocationRate = cp.getElongationTranslocationRate();
		//Event lists
		//this.elongationEventQueue = new PriorityQueue<RibosomeEvent>();
		this.elongationTransferEventQueue = new PriorityQueue<RibosomeEvent>();
		
		 
		this.elongationTranslocationEventQueue = new RibosomeMovementEventQueue();
		this.randomWalkEventQueue = new RibosomeMovementEventQueue();

		
		//elongation binding event list First reaction method
		//elongation binding event list Direct method
		//if(this.isGillespieDirect){
			elongationBindingEventQueue = new ElongationBindingEventQueueDM(tRNAbindingRate, nonGognateIntroduceWaitingTime, nonGognateWaitingTime, nonCognateChangeBindingRate, nonCognateMMConstant, cp.gettRNATotalChargedAbundances(),cp);
			cp.printDebugInfo("the binding of tRNA to the ribosome uses Gillespie Direct Method");
		//} else{
		//	elongationBindingEventQueue = new ElongationBindingEventQueueFR(tRNAbindingRate, nonGognateIntroduceWaitingTime, nonGognateWaitingTime, nonCognateChangeBindingRate, nonCognateMMConstant);
		//	cp.printDebugInfo("the binding of tRNA to the ribosome uses Gillespie FirstReaction Method");
		//}
		
		//tRNA event list
		this.tRNAEventQueue = new PriorityQueue<TRNAEvent>();
		
		bindRibo2mRNAQ = new BindRibo2mRNAQ(cp);
	}
	

	
	
	/**
	 * removes the next elongation event
	 * @return
	 */
	private RibosomeEvent popNextRibosomeElongationTransferEvent(){
		return elongationTransferEventQueue==null?null:elongationTransferEventQueue.poll();
	}
	
	

	
	/**
	 * removes the next elongation event
	 * @return
	 */
	private RibosomeEvent popNextRibosomeElongationTranslocationEvent(){
		return elongationTranslocationEventQueue.poll();
	}
	
		
	
	
	
	
	/**
	 * removes the next elongation event
	 * @return
	 */
	private RibosomeEvent popNextRibosomeRandomWalkEventEvent(){
		return randomWalkEventQueue.poll();
	}
	
	
	
	
	/**
	 * returns the next elongation binding event and removes it from the list
	 * @return
	 */
	private  RibosomeEvent popNextRibosomeElongationBindingEvent(){
		return elongationBindingEventQueue==null?null:elongationBindingEventQueue.poll();
	}
	
	
	
	/**
	 * returns the next elongation binding event and removes it from the list
	 * @return
	 */
	private  TRNAEvent popNexttRNAEvent(){
		return tRNAEventQueue==null?null:tRNAEventQueue.poll();
	}
	


	
	/**
	 * returns the next binding event and removes it from the list
	 * @return
	 */
	public RibosomeEvent popNextRibosomeBindingEvent(){
		return bindRibo2mRNAQ.pop();
	}
	
	/**
	 * returns a number which encodes whether the next event is ribosome binding/ elongation/ tRNA charging
	 * @param stopTime
	 * @return
	 */
	public int getNextEventType(){
		int result = Constants.NEXT_EVENT_IS_RIBOSOME_NONE;
		double nextEventTime =  Double.MAX_VALUE;
		
	 	
		//System.out.println("*************************************");
		//System.out.println(sizeToString());
		if(!bindRibo2mRNAQ.isEmpty() && nextEventTime > bindRibo2mRNAQ.peek().time){
			nextEventTime = bindRibo2mRNAQ.peek().time;
			result = Constants.NEXT_EVENT_IS_RIBOSOME_BINDING;
                        
			//System.out.println("NEXT_EVENT_IS_RIBOSOME_BINDING: "+nextEventTime);
		}
		
		if(!randomWalkEventQueue.isEmpty() && nextEventTime > randomWalkEventQueue.peek().time){
			nextEventTime = randomWalkEventQueue.peek().time;
			result = Constants.NEXT_EVENT_IS_RIBOSOME_RANDOM_WALK;
                       //System.out.println("NEXT_EVENT_IS_RIBOSOME_RANDOM_WALK: "+nextEventTime);
		}
		
		if(!elongationBindingEventQueue.isEmpty() && nextEventTime > elongationBindingEventQueue.peek().time){
			nextEventTime = elongationBindingEventQueue.peek().time;
			result = Constants.NEXT_EVENT_IS_RIBOSOME_ELONGATION_BINDING;
  			
                        //System.out.println("NEXT_EVENT_IS_RIBOSOME_ELONGATION_BINDING: "+nextEventTime+" : "+elongationBindingEventQueue.size());
		}
		
		
		if(!elongationTransferEventQueue.isEmpty() && nextEventTime > elongationTransferEventQueue.peek().time){
			nextEventTime = elongationTransferEventQueue.peek().time;
			result = Constants.NEXT_EVENT_IS_RIBOSOME_ELONGATION_TRANSFER;
		if(elongationTransferEventQueue.size() > 400000)
                    System.out.println("NEXT_EVENT_ elongate transfer: "+nextEventTime + "  " +elongationTransferEventQueue.size());

                        
                        //System.out.println("NEXT_EVENT_IS_RIBOSOME_ELONGATION_TRANSFER: "+nextEventTime);
		}
		
		if(!elongationTranslocationEventQueue.isEmpty() && nextEventTime > elongationTranslocationEventQueue.peek().time){
			nextEventTime = elongationTranslocationEventQueue.peek().time;
			result = Constants.NEXT_EVENT_IS_RIBOSOME_ELONGATION_TRANSLOCATION;
		
     if(elongationTranslocationEventQueue.size() > 400000)
         System.out.println("NEXT_EVENT_ elongate translocation: "+nextEventTime + "  " +elongationTranslocationEventQueue.size());

                        //System.out.println("NEXT_EVENT_IS_RIBOSOME_ELONGATION_TRANSLOCATION: "+nextEventTime);
		}
		
		if(!tRNAEventQueue.isEmpty() && nextEventTime > tRNAEventQueue.peek().time){
			nextEventTime = tRNAEventQueue.peek().time;
			result = Constants.NEXT_EVENT_IS_TRNA_CHARGING;
			//System.out.println("NEXT_EVENT_IS_TRNA_CHARGING: "+nextEventTime);
		}
		
		return result;
	}
	
	/**
	 * returns the soones event
	 * @return
	 */
	public Event getNextEvent(){
		Event e=null;
		int nextEventType = getNextEventType();
		
		//System.out.println("elongation binding: "+elongationEventQueueBinding.size()+"; elongation:"+elongationEventQueue.size()+"; tRNA: "+tRNAEventQueue.size());
		//int size1 = size();
		switch(nextEventType){
			case Constants.NEXT_EVENT_IS_RIBOSOME_BINDING: e =this.popNextRibosomeBindingEvent(); break;
			case Constants.NEXT_EVENT_IS_RIBOSOME_RANDOM_WALK: e=this.popNextRibosomeRandomWalkEventEvent(); break;
			case Constants.NEXT_EVENT_IS_RIBOSOME_ELONGATION_BINDING: e=this.popNextRibosomeElongationBindingEvent(); break;
			case Constants.NEXT_EVENT_IS_RIBOSOME_ELONGATION_TRANSFER: e = this.popNextRibosomeElongationTransferEvent(); break;
			case Constants.NEXT_EVENT_IS_RIBOSOME_ELONGATION_TRANSLOCATION: e = this.popNextRibosomeElongationTranslocationEvent(); break;
			case Constants.NEXT_EVENT_IS_TRNA_CHARGING: e =this.popNexttRNAEvent(); break;
			default: e=null;
		}
		/*if(size1-size()!=1){
			System.out.println(nextEventType+" : "+e);
			System.exit(0);
		}*/
		return e;
	}
	
	/**
	 * checks whether there is any event left in the entire list
	 * @return
	 */
	public boolean isEmpty(){
		if(bindRibo2mRNAQ==null && elongationTransferEventQueue.isEmpty() && elongationTranslocationEventQueue.isEmpty() && randomWalkEventQueue.isEmpty() && elongationBindingEventQueue.isEmpty() && tRNAEventQueue.isEmpty()){
			return true;
			
		}
		return false;
	}
	
	
	/**
	 * computes the total number of events in the lists
	 * @return
	 */
	public int size(){
		int result=0;
		//System.out.println("++++++++++++");


		if(bindRibo2mRNAQ!=null){
			result++;
		}
		//System.out.println(result);

		result+=randomWalkEventQueue.size();
		//System.out.println(randomWalkEventQueue.size());
		result+=elongationBindingEventQueue.size();
		//System.out.println(elongationBindingEventQueue.size());
		result+=elongationTransferEventQueue.size();
		//System.out.println(elongationTransferEventQueue.size());
		result+=elongationTranslocationEventQueue.size();
		//System.out.println(elongationTranslocationEventQueue.size());
		result+=tRNAEventQueue.size();
		//System.out.println(tRNAEventQueue.size());
		//System.out.println("++++++++++++");
		return result;
	}
	
	/**
	 * returns a string with the number of events per each type
	 * @return
	 */
	public String sizeToString(){
		String result="[";
		//System.out.println("++++++++++++");


		if(bindRibo2mRNAQ!=null){
			result+="binding: 1, ";
		} else{
			result+="binding: 0, ";
		}
		//System.out.println(result);

		result+="randomWalk: "+randomWalkEventQueue.size()+", ";
		result+="elongationBinding: "+elongationBindingEventQueue.size()+", ";
		result+="elongationTransfer: "+elongationTransferEventQueue.size()+", ";
		result+="elongationTranslocation: "+elongationTranslocationEventQueue.size()+", ";
		result+="tRNA: "+tRNAEventQueue.size()+"]";
		
		return result;
	}
	
	public String toString(){
		String result="";
		//System.out.println("++++++++++++");


		if(bindRibo2mRNAQ!=null){
			result+="{binding: "+bindRibo2mRNAQ+"}";
		} 
		//System.out.println(result);

		result+="{randomWalk: "+randomWalkEventQueue+"} ";
		result+="{elongationBinding: "+elongationBindingEventQueue+"} ";
		result+="{elongationTransfer: "+elongationTransferEventQueue+"} ";
		result+="{elongationTranslocation: "+elongationTranslocationEventQueue+"} ";
		result+="{tRNA: "+tRNAEventQueue.size()+"}";
		
		return result;
	}
	/**
	 * schedules next tRNA charging... replaces an old one if necessary
	 * @param tRNAID the ID of the tRNA that charges
	 * @param cellTime the time of cell
	 */
	public void scheduletRNACharging(int tRNAID,double cellTime, double chargeRate, Random randomGenerator){
		TRNAEvent te = new TRNAEvent(cellTime+Gillespie.computeNextReactionTime(chargeRate,randomGenerator),tRNAID);
	 
		
		tRNAEventQueue.add(te);
	}
	
	
	
        public void scheduleRandomWalk(int ribosomeID, double cellTime, double IBSmoveRate, Random randomGenerator, int mRNA, double lastReactionTime, Cell cp)
                {
                    
                    scheduleNextRibosomeRandomWalkEvent(ribosomeID, cellTime, IBSmoveRate, randomGenerator, mRNA, lastReactionTime, cp);
                    
                }
        
	/**
	 * inserts a ribosome random walk on mRNA event 
	 * @param ribosomeID the ID of the ribosome
	 * @param cellTime the time of the cell
	 * @param IBSmoveRate the random walk rate
	 * @param randomGenerator the random number generator
	 * @param mRNA the id of the mRNA
	 * @param codonID the ID of the next codon
	 * @param lastReactionTime the time the ribosome performed last action
	 */
	private void scheduleNextRibosomeRandomWalkEvent(int ribosomeID, double cellTime, double IBSmoveRate, Random randomGenerator, int mRNA, double lastReactionTime, Cell cp){
		//double nextTime;
		
		//nextTime = Gillespie.computeNextReactionTime(IBSmoveRate, randomGenerator);
		randomWalkEventQueue.scheduleNextMovement(ribosomeID, cellTime, IBSmoveRate, mRNA, lastReactionTime, cp, randomGenerator);
		//randomWalkEventQueue.add(new RibosomeEvent(cellTime + nextTime,ribosomeID,mRNA,1, -1,-1, codonID,lastReactionTime));
	}
	
	
        
        
        
        
        
        
	/**
	 * schedules a binding of the tRNA to the ribosome event
	 * @param ribosomeID the ID of the ribosome
	 * @param cellTime the current cell time
	 * @param mRNAID the mRNA where the ribosome is bound to
	 * @param currentPosition the current position on the mRNA
	 * @param codonID the ID of the codon
	 * @param randomGenerator the random number generator
	 * @param totalChargedAbundance the total charged abundace
	 * @param tRNApropensity the propensities to bind to the tRNA
	 * @param numberOfFreeBoundRibosomes number of free ribosomes
	 * @param lastReactionTime the time of the last reaction
	 */
	
//            
//            private void scheduleBindingOfTRNAToRibo(int ribosomeID, double cellTime, int mRNAID, int currentPosition, byte codonID, Random randomGenerator, Cell cp, int numberOfFreeBoundRibosomes, double lastReactionTime){
//		this.elongationBindingEventQueue.scheduleNextElongationBindingEvent(ribosomeID, cellTime, mRNAID, currentPosition, codonID, randomGenerator, cp, numberOfFreeBoundRibosomes, lastReactionTime);
//	}

//All this does is to form a public interface for schedule 
        public void scheduleInternalRNATransfer(int ribosomeID, double cellTime, int mRNAID, int currentPosition, int tRNAID, byte codonID, boolean isCognate, Random randomGenerator, double lastReactionTime, Cell cp)   {
    
    scheduleRibosomeElongationTransferEvent(ribosomeID,cellTime,mRNAID,currentPosition,tRNAID, codonID, isCognate, randomGenerator,lastReactionTime,  cp);
		
}
        
        
	/**
	 * schedules a amino-acid transfer event.
	 * @param ribosomeID the ID of the bound ribosome
	 * @param cellTime the current cell time
	 * @param mRNAID the id of the mRNAwhere the ribosome is bound
	 * @param currentPosition the current position of the ribosome on the mRNA
	 * @param tRNAID the ID of the bound tRNA
	 * @param codonID the codon of where the ribosome is bound
	 * @param isCognate true if the tRNA is cognate to the codon
	 * @param randomGenerator the random generator 
	 * @param lastReactionTime the time of the last event performed by current ribosome
	 */
	private void scheduleRibosomeElongationTransferEvent(int ribosomeID, double cellTime, int mRNAID, int currentPosition, int tRNAID, byte codonID, boolean isCognate, Random randomGenerator, double lastReactionTime, Cell cp){
		int nextState = Constants.RIBOSOME_ELONGATION_STATE_TRANSFER;
//Your delay for every attempt no matter what. 
double nextTime;



		double buffer;
		if(isCognate){
			nextTime =Gillespie.computeNextReactionTime(1.0/this.nonGognateWaitingTime,randomGenerator);
			buffer= randomGenerator.nextDouble();
			if(buffer>this.elongationCognateProbability) //the tRNA will unbind again
				nextState = Constants.RIBOSOME_ELONGATION_STATE_EMPTY;
		
                        
                        
		} else{
 nextTime=Gillespie.computeNextReactionTime(1.0/this.nearGognateWaitingTime,randomGenerator);
			buffer= randomGenerator.nextDouble();
			if(buffer>this.elongationNearCognateProbability)//the tRNA will inbind again
				nextState = Constants.RIBOSOME_ELONGATION_STATE_EMPTY;
	

			 
		}
		//cp.printDebugInfo(cellTime+" | "+(cellTime + nextTime)+": ribosomeID = "+ribosomeID + "; nextState = "+nextState);
		
		this.elongationTransferEventQueue.add( new RibosomeEvent(cellTime + nextTime,ribosomeID,mRNAID, currentPosition, nextState , tRNAID, codonID,lastReactionTime));

                
	}

        
        
        
         
        
       
	
	
 
        //this function *is* dispensible....but I leave it for the time being...lazyness
       //it is used exactly once at the moment. 
        
        public void scheduleRibosomeBinding(Cell cp){
            int ribo = cp.getLastFreeRibosome();
            
            if(ribo>-1){
                      this.bindRibo2mRNAQ.updateRibosomeBindingPropensities( cp); 
                      
 scheduleNextRibosomeBindingEvent(ribo  , cp.getCellTime(),   cp.getNoOfFreeRibosomes(), cp, this.bindRibo2mRNAQ.ribosomeBindingPropensitySum, this.bindRibo2mRNAQ.ribosomeBindingPropensity, cp.getRandomGenerator(),cp.getRibosomeLastReactionTime(ribo));
            }else{Ribosome.setIsFreeRibos(false);}
        
        
        }
        
        
        
        
        
        
       public void scheduleRibosomeBinding(Cell cp,int mRNAtype){
           int ribo = cp.getLastFreeRibosome();
            
           
    
           
            if(ribo>-1){
           
           
           // System.out.println("Schedule Ribo Binding" + ribo );
          //update the binding propensities
                        this.bindRibo2mRNAQ.updateRibosomeBindingPropensities(mRNAtype,cp);    
 scheduleNextRibosomeBindingEvent(ribo  , cp.getCellTime(),   cp.getNoOfFreeRibosomes(), cp, this.bindRibo2mRNAQ.ribosomeBindingPropensitySum, this.bindRibo2mRNAQ.ribosomeBindingPropensity, cp.getRandomGenerator(),cp.getRibosomeLastReactionTime(ribo));
            }else{Ribosome.setIsFreeRibos(false);}
           
        }  
        
        
        
        
	
	/**
	 * schedules the next ribosome binding event
	 */
	private boolean scheduleNextRibosomeBindingEvent(int ribosomeID, double cellTime,  int freeRibosomes, Cell cp, double ribosomeBindingPropensitySum,double[] ribosomeBindingPropensity, Random randomGenerator, double lastReactionTime){
		
        
		
		// if no ribosome then halt
		if(freeRibosomes>0 && ribosomeBindingPropensitySum>0){ 
                 
                     
			//generate next reaction time
			double nextTime = Gillespie.computeNextReactionTime(ribosomeBindingPropensitySum*freeRibosomes, randomGenerator);
			
			//find next reaction
                        //***NOTE*** It is not necessary (and indeed leads to an error) to multiply the propensity argument by the number of free Ribos!
			int nextmRNAType = Gillespie.getNextReaction(ribosomeBindingPropensitySum*randomGenerator.nextDouble(), ribosomeBindingPropensity);
		
                         
                        if(nextmRNAType==-1){
                        //if we are here this means that we suffered from a rounding error. Fix that:
                            
                         bindRibo2mRNAQ.updateRibosomeBindingPropensities(-21,cp);   
                         
                            
                          return false; 
                        }
                        
                        int lastmRNA =cp.getRandomFreemRNA(nextmRNAType);
			byte codonID = -1;
			this.bindRibo2mRNAQ.add(new RibosomeEvent(cellTime+nextTime,ribosomeID,lastmRNA, 0, 0,-1,codonID,lastReactionTime));
                        
                       
                        return true;
		}
                
                return false; 
	}
	
	/**
	 * updates the RibsomeBinding Propensities
	 * @param gotFree
	 * @param cp
	 */
	public void updateRibosomeBindingPropensities(FreedAgents gotFree, Cell cp){
		
	}
	
	/**
	 * returns true if the ribosome binding event is empty 
	 * @return
	 */
	public boolean isRibosomeBindingEventQueueEmpty(){
		return this.bindRibo2mRNAQ.isEmpty();
	}
	
	/**
	 * removes the binding event
	 */
//	public void clearRibosomeBindingEventQueue(){
//		this.bindRibo2mRNAQ.clear();
//	}
//        
        
        
        public    BindRibo2mRNAQ    getRibo2MRNAEventQ()
        {
            
            return this.bindRibo2mRNAQ;
        }
        
        
        public   ElongationBindingEventQueueDM gettRNABindingQ(){
            return elongationBindingEventQueue;
        }
        
        
        public RibosomeMovementEventQueue getTranslocationQ(){
            
            return this.elongationTranslocationEventQueue;
        }
        
}
