package event;

import utils.Constants;

public class RibosomeEvent extends Event {
	
	public int ribosomeID;
	public int mRNAID;
	public int position;
	public int state;
	public int tRNAID;
	public byte codonID;
	public double lastReactionTime;
	
	public RibosomeEvent(double time, int ribosomeID,int mRNAID,int position,int state, int tRNAID, byte codonID, double lastReactionTime){
		super(time);
		this.ribosomeID = ribosomeID;
		this.mRNAID = mRNAID;
		this.position = position;
		this.state = state;
		this.tRNAID=tRNAID;
		this.codonID = codonID;
		this.lastReactionTime = lastReactionTime;
	}
	
	public boolean isBindingEvent(){
		return mRNAID>=0 && position==0;
	}
	
	public String toString(){
		String stateStr="";
		if(state==Constants.RIBOSOME_ELONGATION_STATE_EMPTY){
			stateStr = " with no tRNA bound";
		} else if(state==Constants.RIBOSOME_ELONGATION_STATE_BOUND){
			stateStr = " with a tRNA of type t"+this.tRNAID+" bound";
		} else if(state==Constants.RIBOSOME_ELONGATION_STATE_TRANSFER){
			stateStr = " with a tRNA of type t"+this.tRNAID+" bound and the amino-acid transfered";
		}
		return time+": ribosome " + ribosomeID + " on mRNA "+ mRNAID + " at position "+ position + " "+ stateStr;
	}
	
	
	public boolean isEqualTo(RibosomeEvent re){
		return this.ribosomeID==re.ribosomeID && this.time == re.time;
	}
}
