package event;

import java.util.PriorityQueue;

import utils.Constants;
import utils.Gillespie;
import java.util.Random;
import environment.Cell;

public class RibosomeMovementEventQueue {
	private PriorityQueue<RibosomeEvent> eventQueue;
	
	
	
	public RibosomeMovementEventQueue(){
		eventQueue=new PriorityQueue<RibosomeEvent>();
	}
	
	/**
	 * returns the next translocation event
	 * @return
	 */
	public RibosomeEvent peek(){
		return eventQueue.peek();
	}
	
	/**
	 * returns the next translocationt and removes it from the list
	 * @return
	 */
	public RibosomeEvent poll(){
		return eventQueue.poll();
	}
	
	/**
	 * adds a new translocation event
	 * @param re the new event
	 */
	public void add(RibosomeEvent re){
		eventQueue.add(re);
	}
	
	
	/**
	 * returns true if the binding event is null
	 * @return
	 */
	public boolean isEmpty(){
		return eventQueue.isEmpty();
	}
	
        
        
        
        //This is in essence the same as scheduleNextMovement.
        //Only that it does not schedule anything, but instead returns the event
        
        
        public RibosomeEvent  getMoveOnEvent(RibosomeEvent e,  Cell cp){
		
          int  ribosomeID=e.ribosomeID   ;
          double  cellTime=cp.getCellTime();
          double  moveRate=cp.getElongationTranslocationRate();
          int  mRNAID = e.mRNAID;
            
       
                RibosomeEvent riboEv=null ;
	//even if we turn off moving, we still need it in case we get stuck behind sbd. 
        //then we use the negative moveRate
  if(moveRate >0){System.out.println("We should not be in getMobveOnEvent");System.exit(-1);}
                
		int nextPos;
		if(cp.doesStallEventsWait()){
			nextPos = cp.canMove(ribosomeID);
		}
		nextPos = cp.getRibosomeNextmRNAPosition(ribosomeID);
		
		// || !cp.unbindsFromCodonIfBusy() || !(cp.unbindsFromIBSifFree()&&cp.isAtIBSSite(ribosomeID))|| !(cp.unbindsFromterminationIfFree()&&cp.isAtTerminationSite(ribosomeID))
		
		if(cp.getBlockedBy(ribosomeID) == Constants.NONE){
			
			
			byte codonID =Constants.NONE;
			if(nextPos!=Constants.NONE){
				codonID=cp.getCodon(mRNAID, nextPos);
			}
			//cp.printDebugInfo("ribosome = "+ribosomeID + "; nextPos = "+nextPos+"; nextCodon:"+codonID);
			
riboEv=new RibosomeEvent(cellTime ,ribosomeID,mRNAID,nextPos,  Constants.RIBOSOME_ELONGATION_STATE_EMPTY,-1, codonID,cp.getCellTime());
                        

			
			//if(cp.getBlockedBy(ribosomeID))
			///cp.setBlocks(cp.getBlockedBy(ribosomeID),Constants.NONE);
			//cp.setBlockedBy(ribosomeID, Constants.NONE);
		}
		
		return riboEv;
	}
        
        
        
        
        
        
        
	/**
	 * schedules the next movement of the ribosome on the mRNA
	 * @param ribosomeID the id of the ribosome to move
	 * @param cellTime the current time
	 * @param moveRate the rate at which the ribosome moves
	 * @param mRNAID the id of the mRNA where it is bound
	 * @param codonID the codonID of the next position 
	 * @param lastReactionTime
	 * @param cp
	 * @param randomGenerator
	 * @return
	 */
	public boolean  scheduleNextMovement(int ribosomeID, double cellTime, double moveRate, int mRNAID, double lastReactionTime, Cell cp, Random randomGenerator){
		boolean result = false;
	//even if we turn off moving, we still need it in case we get stuck behind sbd. 
        //then we use the negative moveRate
                if(moveRate <0) moveRate*=-1;
                
		int nextPos;
		if(cp.doesStallEventsWait()){
			nextPos = cp.canMove(ribosomeID);
		}
		nextPos = cp.getRibosomeNextmRNAPosition(ribosomeID);
		
		// || !cp.unbindsFromCodonIfBusy() || !(cp.unbindsFromIBSifFree()&&cp.isAtIBSSite(ribosomeID))|| !(cp.unbindsFromterminationIfFree()&&cp.isAtTerminationSite(ribosomeID))
		
		if(cp.getBlockedBy(ribosomeID) == Constants.NONE){
			double nextTime = Gillespie.computeNextReactionTime(moveRate, randomGenerator);
			
			byte codonID =Constants.NONE;
			if(nextPos!=Constants.NONE){
				codonID=cp.getCodon(mRNAID, nextPos);
			}
			//cp.printDebugInfo("ribosome = "+ribosomeID + "; nextPos = "+nextPos+"; nextCodon:"+codonID);
			
			add(new RibosomeEvent(cellTime + nextTime,ribosomeID,mRNAID,nextPos,  Constants.RIBOSOME_ELONGATION_STATE_EMPTY,-1, codonID,lastReactionTime));
			result = true;
			
			//if(cp.getBlockedBy(ribosomeID))
			///cp.setBlocks(cp.getBlockedBy(ribosomeID),Constants.NONE);
			//cp.setBlockedBy(ribosomeID, Constants.NONE);
		}
		
		return result;
	}
	
	/**
	 * returns size
	 * @return
	 */
	public int size(){
		return this.eventQueue.size();
	}
	
}
