package objects;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.StringTokenizer;

import utils.Constants;

public class Codons {
	// list of all codons
	public String[] codon = {"AAA", "AAC", "AAG", "AAU", "ACA", "ACC", "ACG", "ACU", "AGA", "AGC", "AGG", "AGU", "AUA", "AUC", "AUU", "CAA", "CAC", "CAG", "CAU", "CCA", "CCC", "CCG", "CCU", "CGA", "CGC", "CGG", "CGU", "CUA", "CUC", "CUG", "CUU", "GAA", "GAC", "GAG", "GAU", "GCA", "GCC", "GCG", "GCU", "GGA", "GGC", "GGG", "GGU", "GUA", "GUC", "GUG", "GUU", "UAC", "UAU", "UCA", "UCC", "UCG", "UCU", "UGC", "UGG", "UGU", "UUA", "UUC", "UUG", "UUU", "AUG", "UAG", "UGA","UAA", "IBS"};  
	
	
	
	//list of the corresponding amino-acid
	public String[] aminoAcid = {"F", "F", "L", "L", "S", "S", "S", "S", "Y", "Y", "C", "C", "W", "L", "L", "L", "L", "P", "P", "P", "P", "H", "H", "Q", "Q", "R", "R", "R", "R", "I", "I", "I", "M", "T", "T", "T", "T", "N", "N", "K", "K", "S", "S", "R", "R", "V", "V", "V", "V", "A", "A", "A", "A", "D", "D", "E", "E", "G", "G", "G", "G", "M",  "STOP", "STOP", "STOP", "INIT"};  
	
	
	

	//list of mapping of codon to a number
	public HashMap<String, Byte> codonId;
	
	
	//list of start codons
	public byte[] startCodon;
	
	//list of stop codons
	public byte[] stopCodon;
 
	
	public HashMap<String,String> aminoAcids = getAminoAcidNames(Constants.AMINO_ACIDS_FILENAME);
	
	public HashMap<String, Byte> buildCodonId(String[] codon){
		HashMap<String, Byte> ids = new HashMap<String, Byte>();
		
		for (int i= 0; i< codon.length; i++){
			ids.put(codon[i], (byte) i);
		}
		
		return ids;
	}
	
	public Codons(){
		codonId=buildCodonId(codon);

		//System.out.println(codon.length);
		//System.out.println(codonId.size());
		//System.out.println(aminoAcid.length);
		
		startCodon = new byte[2];
		startCodon[0]=codonId.get("IBS"); 
		startCodon[1]=codonId.get("AUG");
		stopCodon = new byte[3];
		stopCodon[0] = codonId.get("UAG");
		stopCodon[1] = codonId.get("UGA"); 
		stopCodon[2] = codonId.get("UAA");
	}
	
	
	private HashMap<String,String> getAminoAcidNames(String filename){
		HashMap<String,String> results = new HashMap<String,String>();

        BufferedReader reader = null;
		
        String shortName,longName;
        
		try
        {
            reader = new BufferedReader(new FileReader(filename));
            String text = null;
            StringTokenizer tempStringTokenizer;
         

            // repeat until all lines is read
            while ((text = reader.readLine()) != null)
            {
            		longName="";
            		shortName="";
  	                // this statement will break the string into the words which are separated by space.
	            	tempStringTokenizer = new StringTokenizer(text,",");

	                // push all the words to the stack one by one
	                if (tempStringTokenizer.hasMoreTokens()) {
	                	shortName = tempStringTokenizer.nextElement().toString();
	                }
	                if (tempStringTokenizer.hasMoreTokens()) {
	                	longName = tempStringTokenizer.nextElement().toString();
	                }
	                results.put(longName, shortName);
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            try
            {
                if (reader != null){
                    reader.close();
                }
            } catch (IOException e)
            {
                e.printStackTrace();
            }
        }
		
		
		return results;
	}
	
	/**
	 * cheks whether a specified codon is a termination codon or not
	 * @param codon the string representing the codon
	 * @return true if this is a termination codon or false otherwise
	 */
	public boolean isStopCodon(String codon){
		if(codonId.containsKey(codon)){
			return isStopCodon(codonId.get(codon));
		}
		return false;
	}
	
	/**
	 * 
	 * @param codonID
	 * @return
	 */
	public boolean isStopCodon(byte codonID){
		for(byte i:stopCodon){
			if(i==codonID){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * 
	 * @param codonID
	 * @return
	 */
	public boolean isSartCodon(byte codonID){
		for(byte i:startCodon){
			if(i==codonID){
				return true;
			}
		}
		return false;
	}
}
