package objects;

import utils.Constants;

public class MessengerRNA {

	private int[] occupancy;
	private String name;
	private int type;
	
	public MessengerRNA(String name, int type, int length){
		this.name =name;
		this.occupancy = new int[length];
		for(int i=0;i<length;i++){
			this.occupancy[i]=Constants.NONE;
		}
		this.type=type;
	}
	
	/**
	 * mutator method that returns the type of the mRNA strand
	 * @return the type of mRNA strand
	 */
	public String getName(){
		return name;
	}
	
	/**
	 * checks whether a position on the strand is free or not
	 * @param position the position on the strand
	 * @return the ID of the occuping ribosome (-1 if free)
	 */
	public int isOccupied(int position){
		if(position>=0 && position<occupancy.length){
			return occupancy[position];
		} else{
			return Constants.NONE;
		}
	}
	
	/**
	 * checks whether a range of sites are free
	 * @param start the start position of the range (inclusive)
	 * @param end the end position of the range (inclusive)
	 * @return  the ID of the occuping ribosome (-1 if free)
	 */
	public int areOccupied(int start, int end){

		//for(int i=start;i<=end;i++){System.out.print(occupancy[i]);}System.out.println(start+" < "+end);
		
		if(start >=0 && end >= start && end<occupancy.length){
			for(int i=start;i<=end;i++){
				if(occupancy[i] >=0){
					return occupancy[i];
				}
			}
		} 
		
		return Constants.NONE;
	}
	
	
	/**
	 * gets the id of the mRNA type. It speeds the search of the mRNA type
	 * @return the mRNA type id
	 */
	public int getType(){
		return type;
	}
	
	/**
	 * changes the id of the mRNA type
	 * @param id
	 */
	public void setType(int type){
		this.type=type;
	}
	
	
	/**
	 * returns the length of the mRNA strand
	 * @return 
	 */
	public int getLength(){
		return occupancy.length;
	}
	

	/**
	 * sets the mRNA as occupied/free at position pos
	 * @param pos the position where mRNA gets occupied
	 * @param occupancyState the ID of the occuping ribosome -1 if none
	 */
	public void setOccupancyState(int pos, int ribosomeID){
		if(pos>=0 && pos<occupancy.length){
			occupancy[pos]= ribosomeID;
		}
	}
	
	
	/**
	 * sets the mRNA as occupied/free between positions start and end (inclusive)
	 * @param start the start postion (inclusive)
	 * @param end the end postion (inclusive)
	 * @param occupancyState the ID of the occuping ribosome -1 if none
	 */
	public void setOccupancyState(int start, int end, int occupancyState){
		if(start<=end){
			if(start<0){ start=0;}
			if(start>=occupancy.length){start =	occupancy.length-1;}
			if(end<0){ end=0;}
			if(end>=occupancy.length){end=occupancy.length-1;}

			for(int i=start; i<=end;i++){
				occupancy[i]=occupancyState;
			}
		}
	}
	
	/**
	 * moves one position to the left
	 * @param currentPos
	 * @param nextPos
	 */
	public void slide(int currentPos, int nextPos){
		occupancy[nextPos] = occupancy[currentPos];
		occupancy[currentPos]=Constants.NONE;
	}
	
	
	/*public int getFirstRibosome(int pos){
		int result;
		for(int i=pos; i<occupancy.length;i++){
			if (occupancy[i]!=Constants.NONE){}
		}
	}*/
	
}
