package objects;


/**
 * class which stores tRNA information
 * @author nrz2
 *
 */
public class TranslationalRNA {
	
	private byte label; // the id of the tRNA
	private String aminoAcid; // the Amino Acid
	private int abundance;
	private int charged;
	private int busy;
	private int uncharged;
	private double rechargeRate;
	
	/**
	 * class constructor
	 * @param id the id of the tRNA
	 * @param aminoAcid the id of the amino acid
	 * @param name the name of the aminoacid
	 */
	public TranslationalRNA(byte label, String aminoAcid, int abundance, double rechargeRate){
		this.label = label;
		this.aminoAcid = aminoAcid;
		this.abundance=abundance;
		this.charged=abundance;
		this.busy = 0;
		this.uncharged = 0;
		this.rechargeRate = rechargeRate;
	}
	
	/**
	 * class constructor
	 * @param id the id of the tRNA
	 * @param aminoAcid the id of the amino acid
	 * @param name the name of the aminoacid
	 */
	public TranslationalRNA(TranslationalRNA tRNA){
		this.label = tRNA.label;
		this.aminoAcid = tRNA.aminoAcid;
		this.abundance=tRNA.abundance;
		this.charged=tRNA.abundance;
		this.busy = tRNA.busy;
		this.uncharged = tRNA.uncharged;
		this.rechargeRate = tRNA.rechargeRate;
	}
	
	/**
	 * updates the recharge rate
	 * @param rechargeRate
	 */
	public void setRechargeRate(double rechargeRate){
		this.rechargeRate = rechargeRate;
	}
	
	/**
	 * returns the reacharge rate
	 * @return
	 */
	public double getRechargeRate(){
		return rechargeRate;
	}
	
	/**
	 * returns true if the codon supplied as a parameter is cognate or false otherwise
	 * @param codon
	 * @return
	 */
	/*public boolean isCognate(byte codon){
		for(int i=0;i<cognateCodons.size();i++){
			if(cognateCodons.get(i) == codon){
				return true;
			}
		}
		return false;
	}*/
	
	
	/**
	 * returns true if the codon supplied as a parameter is near cognate or false otherwise
	 * @param codon
	 * @return
	 */
	/*public boolean isNearCognate(byte codon){
		for(int i=0;i<nearCognateCodons.size();i++){
			if(nearCognateCodons.get(i) == codon){
				return true;
			}
		}
		return false;
	}*/
	
	
	
	
	/**
	 * returns true if the codon supplied as a parameter is non cognate or false otherwise
	 * @param codon
	 * @return
	 */
	/*public boolean isNonCognate(byte codon){
		// check if it is non cognate
		for(int i=0;i<cognateCodons.size();i++){
			if(cognateCodons.get(i) == codon){
				return false;
			}
		}

		// check if it is near cognate
		for(int i=0;i<nearCognateCodons.size();i++){
			if(nearCognateCodons.get(i) == codon){
				return false;
			}
		}
		return true;
	}*/
	
	/**
	 * return the id of the tRNA type
	 * @return
	 */
	public byte getLabel(){
		return label;
	}
	
	/**
	 * return the aminoAcid of the tRNA type
	 * @return
	 */
	public String getAminoAcid(){
		return aminoAcid;
	}
	
	
	
	/**
	 * return the list of cognate codons
	 * @return
	 */
	/*public ArrayList<Integer> getCognateCodons(){
		return cognateCodons;
	}*/
	
	/**
	 * return the list of near cognate codons
	 * @return
	 */
	/*public ArrayList<Integer> getNearCognateCodons(){
		return nearCognateCodons;
	}*/
	
	/**
	 * sets the abundance of the tRNA
	 * @param abundance
	 */
	public void setAbundance(int abundance){
		this.abundance=abundance;
	}
	
	/**
	 * returns the abundance of this tRNA
	 * @return
	 */
	public int getAbundance(){
		 return abundance;
	}
	
	/**
	 * generates a string with the data stored by the object
	 * @return 
	 */
	public String toString(){
		String result="";
		result +=label+"("+aminoAcid+"): " + abundance;
		return result;
	}
	
	/**
	 * returns the abundance of charged tRNAs
	 * @return
	 */
	public int getChargedAbundance(){
		return this.charged;
	}
	
	
	
	/**
	 * returns the abundance of busy tRNAs
	 * @return
	 */
	public int getBusyAbundance(){
		return this.busy;
	}
	
	
	
	
	/**
	 * returns the abundance of uncharged tRNAs
	 * @return
	 */
	public int getUnchargedAbundance(){
		return this.uncharged;
	}
	
	
	
	
	/**
	 * sets a charged tRNA to an busy one (bound to the ribosome)
	 * @return true if the charged abundance is higher than 0 or false otherwise
	 */
	public boolean setBusy(){
		if(this.charged==0){
			return false;
		}
		
		this.charged--;
		this.busy++;
		return true;
	}
	
	
	/**
	 * sets a busy (bound to the ribosome) tRNA to an charged one 
	 * @return true if the busy abundance is higher than 0 or false otherwise
	 */
	public boolean resetBusy(){
		if(this.busy==0){
			return false;
		}
		
		this.charged++;
		this.busy--;
          
                
		return true;
	}
	
	
	/**
	 * sets a busy (bound to the ribosome) tRNA to an uncharged one 
	 * @return true if the busy abundance is higher than 0 or false otherwise
	 */
	public boolean uncharge(){
           
            
		if(this.busy==0){
			return false;
		}
		
		this.uncharged++;
		this.busy--;
		return true;
	}
	
	
	/**
	 * sets an  uncharged  tRNA to an charged one 
	 * @return true if the uncharged abundance is higher than 0 or false otherwise
	 */
	public boolean charge(){
		if(this.uncharged==0){
			return false;
		}
		
		this.charged++;
		this.uncharged--;
		return true;
	}
	
}
