package utils;

import environment.Cell;
import java.util.Random;
import objects.Codons;

/**
 * class that generates random codons and can mad a codon sequence to a number
 * @author nrz2
 *
 */

public class SequenceUtils {

	public static final Codons codons=new Codons();

	//public static final TranslationalRNAstaticTable tRNAtable = new TranslationalRNAstaticTable(Constants.TRNA_ABUNDANCE_FILENAME,Constants.TRNA_CHARGE_RATE);
	public static final TranslationalRNAstaticTable tRNAtable = new TranslationalRNAstaticTable(Cell.tRNAAbundaceFile,Constants.TRNA_CHARGE_RATE);
	
	/**
	 * generates a random codon sequence for an mRNA which starts with AUG and ends with a random sequence between (UAG, UGA, UAA)
	 * note that the first position is -1 which represents the initiation site  
	 * @param length the length of the mRNA in number of codons
	 * @param generator the random number generator
	 * @return the codon sequence of the mRNA
	 */
	public static byte[] randomMessengerRNA(int length, Random generator){
		byte[] sequence = new byte[length+1];
		//System.out.println(length);
		for(int i=0;i<length;i++){
			sequence[i]=0;
		}
		sequence[0]=codons.startCodon[0];
		sequence[1]=codons.startCodon[1];
		//Random generator = new Random();
		for(int i=2;i<length;i++){
			sequence[i] = (byte) generator.nextInt(codons.codon.length-5);
			/*sequence[i] = (byte)generatePoissonRandomNumber(generator,Constants.POISSON_MEAN);
			if(sequence[i]>(codons.codon.length-5)){
				sequence[i] =(byte) ((byte) codons.codon.length-5);
			}*/
		}
		

		sequence[length] = codons.stopCodon[generator.nextInt(codons.stopCodon.length)];
		
	
		return sequence;
	}
	
	
	/**
	 * transforms a string of characters into a 
	 * @param text
	 * @return
	 */
	public static byte[] messengerRNATrand(String text){
		
		text="IBS"+text;
		if(Math.ceil(text.length()/3)!=Math.floor(text.length()/3)){
			System.out.println("error: string length is not a multiple of 3; "+text.length());
		}
		
		int length =text.length()/3;
		byte[] sequence = new byte[length];
		String str = "";
		
		for(int i=0;i<sequence.length;i++){
			str = "";
			str = text.substring(i*Constants.CODON_LENGTH, (i+1)*Constants.CODON_LENGTH);
			str = str.toUpperCase().replaceAll("T", "U");
			//System.out.println(str);
			if (codons.codonId.containsKey(str)){
				sequence[i] = codons.codonId.get(str);
			} else{
				System.err.println("Error reading mRNA file");
			}
		}
		
		return sequence;
	}
	
	/**
	 * counts the number of termination codons
	 * @param sequence the mRNA sequence
	 * @return the number of termination codons
	 */
	public static int countTerminationSites(byte[] sequence){
		int result=0;
		for(int i=0;i<sequence.length;i++){
			if(codons.isStopCodon(sequence[i])){
				result++;
			}
		}
		return result;
	}
	
	/**
	 * counts how many start codons are on a sequence
	 * @param sequence
	 * @return
	 */
	public static int countStartSites(byte[] sequence){
		int result=0;
		for(int i=0;i<sequence.length;i++){
			if(codons.isSartCodon(sequence[i])){
				result++;
			}
		}
		return result;
	}
	
	/**
	 * method to generate a poissone random number
	 * @param generator the random number generator
	 * @param mean the mean
	 * @return
	 */
	public static int generatePoissonRandomNumber(Random generator, double mean){
		double l = Math.exp(-mean);
		int k=0;
		double p=1;
		while (p>l){
			k++;
			p=p*generator.nextDouble();
		}
		return k-1;

	}
	
	
	/**
	 * returns the relative mRNA occupancy by ribosomes
	 * @param count the number of ribosomes on the mRNA
	 * @param length the length in codons of the mRNA
	 * @param copyNumber the copy number of the gene
	 * @param sitesAfter the number of codons occupied by the ribosome in front of the ribosome
	 * @param sitesBefore the number of codons occupied by the ribosome in back of the ribosome
	 * @return
	 */
	public static double getRelativeRibosomeOccupancy(double count, int length,int copyNumber, int sitesAfter, int sitesBefore){
		double result = 0;
		int totalNoOfPositions =  getTotalNoOfPositions(length, sitesAfter, sitesBefore);
		result = (double)count/(totalNoOfPositions*copyNumber);
		return result;
	}
	
	/**
	 * returns the total number of positions on the mRNA that ribsome can occupy
	 * @param length the length in codons
	 * @param sitesAfter the number of codons occupied by the ribosome in front of the ribosome
	 * @param sitesBefore the number of codons occupied by the ribosome in back of the ribosome
	 * @return
	 */
	public static int getTotalNoOfPositions(int length, int sitesAfter, int sitesBefore){
		//3 comes from 1 IBS, 1 first position and 1 last position
		return  (int) (3 + Math.floor(length-(3+sitesAfter + sitesBefore))/(sitesAfter + sitesBefore + 1));
	}
	
}
