package utils;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

import objects.TranslationalRNA;

public class TranslationalRNAstaticTable {
	//private Codon codons;
	private TranslationalRNA[] tRNAs;
	private ArrayList<ArrayList<Integer>> cognatetRNAs;
	private ArrayList<ArrayList<Integer>> nearCognatetRNAs;
	private int totalAbundance;
	private int chargedAbundance;
	private int busyAbundance;	
	private int unchargedAbundance;
	private boolean[][] isCognate,isNearCognate;
	
	private ArrayList<ArrayList<Integer>> cognateCodons;
	private ArrayList<ArrayList<Integer>> nearCognateCodons;
	private ArrayList<ArrayList<Integer>> affectedCodons;
	
	/**
	 * constructor that initialises the table from a file
	 * @param filename
	 * @param tRNAchargeRate
	 */
	public TranslationalRNAstaticTable(String filename, double tRNAchargeRate){
		initialiseParameters();
		this.loadFile(filename, tRNAchargeRate);
	}
	
	/**
	 * empty class constructor
	 */
	public TranslationalRNAstaticTable(){
		initialiseParameters();
	}

	/**
	 * initialises all the parameters of the class
	 */
	private void initialiseParameters(){
		tRNAs = new TranslationalRNA[0];
		int length = SequenceUtils.codons.codon.length;
		cognatetRNAs = new ArrayList<ArrayList<Integer>>();
		nearCognatetRNAs = new ArrayList<ArrayList<Integer>>();
		for(int i=0;i<length;i++){
			cognatetRNAs.add(new ArrayList<Integer>());nearCognatetRNAs.add(new ArrayList<Integer>());
		}
		
		cognateCodons = new ArrayList<ArrayList<Integer>>();
		nearCognateCodons = new ArrayList<ArrayList<Integer>>();
		affectedCodons= new ArrayList<ArrayList<Integer>>();
		totalAbundance = 0;
		this.chargedAbundance = 0;
		this.busyAbundance = 0;
		this.unchargedAbundance = 0;
		
		isCognate = null;
		isNearCognate = null;
	}
	
	/**
	 * load the data from file
	 * @param filename the filename
	 * @param tRNAchargeRate the rate at which the tRNA gets charged
	 */
	public void loadFile(String filename, double tRNAchargeRate){
        BufferedReader reader = null;
        try
        {
            reader = new BufferedReader(new FileReader(filename));
            String text = null;
            String bufferText;
            StringTokenizer tempStringTokenizer;
            ArrayList<String> buffer;
            ArrayList<TranslationalRNA> tRNAs = new ArrayList<TranslationalRNA>();
            // repeat until all lines is read
            while ((text = reader.readLine()) != null)
            {
            	
            	if(!text.startsWith("#")){
  	                // this statement will break the string into the words which are separated by space.
	            	tempStringTokenizer = new StringTokenizer(text,",");
	            	buffer=new ArrayList<String>();
	                // push all the words to the stack one by one
	                while(tempStringTokenizer.hasMoreTokens()) {
	                	buffer.add(tempStringTokenizer.nextElement().toString());
	                }
	                if(buffer.size() >= 6){
	                	//System.out.println(buffer.get(0));
		                byte codonID = SequenceUtils.codons.codonId.get(buffer.get(0));
		                //compute cognate tRNAs for a codon
		                tempStringTokenizer = new StringTokenizer(buffer.get(3),"; ");
		                while(tempStringTokenizer.hasMoreTokens()) {
		                	bufferText=tempStringTokenizer.nextElement().toString().replaceAll("t", "");
		                	//System.out.println("id:"+bufferText+" | "+codonID);
		                	cognatetRNAs.get(codonID).add(Integer.parseInt(bufferText));
		                }
		                //compute near cognate tRNAs for a codon
		                tempStringTokenizer = new StringTokenizer(buffer.get(5),"; ");
		                while(tempStringTokenizer.hasMoreTokens()) {
		                	bufferText=tempStringTokenizer.nextElement().toString().replaceAll("t", "");
		                	nearCognatetRNAs.get(codonID).add(Integer.parseInt(bufferText));
		                }
		                
		                int abundance = Integer.parseInt(buffer.get(4))/cognatetRNAs.get(codonID).size();
		                byte id;
		                for(int label:cognatetRNAs.get(codonID)){
		                	id = (byte) gettRNA(tRNAs,(byte)label);
		                	if(id <0 ){
		                		TranslationalRNA tRNA = new TranslationalRNA((byte)label,buffer.get(2),abundance,tRNAchargeRate);
		                		tRNAs.add(tRNA);
		                	}
		                }
	                }
	                
            	}
            }
            
    		//this.tRNAs = new TranslationalRNA[Constants.NO_OF_TRNAS];
            this.tRNAs = new TranslationalRNA[maxtRNALabel(tRNAs)+1];
    		for(int i=0;i<this.tRNAs.length;i++){
    			this.tRNAs[i] = new TranslationalRNA((byte)i,"",0,tRNAchargeRate);
    		}
    		totalAbundance = 0;
    		for(TranslationalRNA tRNA:tRNAs){
    			this.tRNAs[tRNA.getLabel()] = tRNA;
    			totalAbundance+=tRNA.getAbundance();
    		}
    		this.chargedAbundance = totalAbundance;
    		
    		
    		//build isCognate/isNearCognate matrices
    		isCognate = new boolean[SequenceUtils.codons.codon.length][this.tRNAs.length];
    		isNearCognate = new boolean[SequenceUtils.codons.codon.length][this.tRNAs.length];
    		for(int i=0;i<isCognate.length;i++){
    			for(int j=0;j<isCognate[i].length;j++){
    				isCognate[i][j] = false;
    				isNearCognate[i][j] = false;
    			}
    		}
    		for(int i=0; i< cognatetRNAs.size();i++){
    			for(int j=0; j< cognatetRNAs.get(i).size();j++){
    				isCognate[i][cognatetRNAs.get(i).get(j)]=true;
    			}
    		}
    		for(int i=0; i< nearCognatetRNAs.size();i++){
    			for(int j=0; j< nearCognatetRNAs.get(i).size();j++){
    				isNearCognate[i][nearCognatetRNAs.get(i).get(j)]=true;
    			}
    		}
    		
    		//build (near)cognateCodons
    		affectedCodons= new ArrayList<ArrayList<Integer>>();
    		cognateCodons = new ArrayList<ArrayList<Integer>>();
    		nearCognateCodons = new ArrayList<ArrayList<Integer>>();
    		for(int i=0;i<this.tRNAs.length;i++){
    			affectedCodons.add(new ArrayList<Integer>());
    			cognateCodons.add(new ArrayList<Integer>());
    			nearCognateCodons.add(new ArrayList<Integer>());
    		}
    		for(int i=0;i<cognatetRNAs.size();i++){
    			for(int j=0;j<cognatetRNAs.get(i).size();j++){
    				//System.out.println(i+" : "+j+" : "+cognatetRNAs.get(i).get(j));
    				cognateCodons.get(cognatetRNAs.get(i).get(j)).add(i);
    				affectedCodons.get(cognatetRNAs.get(i).get(j)).add(i);
    			}
    		}
    		for(int i=0;i<cognatetRNAs.size();i++){
    			for(int j=0;j<nearCognatetRNAs.get(i).size();j++){
    				//System.out.println(i+" : "+j+" : "+nearCognatetRNAs.get(i).get(j));
    				nearCognateCodons.get(nearCognatetRNAs.get(i).get(j)).add(i);
    				affectedCodons.get(nearCognatetRNAs.get(i).get(j)).add(i);
    			}
    		}
            
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            try
            {
                if (reader != null){
                    reader.close();
                }
            } catch (IOException e)
            {
                e.printStackTrace();
            }
        }
	}
	
	
	/**
	 * returns a list with the affected codons by a tRNA species
	 * @param tRNAID
	 * @return
	 */
	public ArrayList<Integer> getAffectedCodons(int tRNAID){
		return affectedCodons.get(tRNAID);
	}
	
	
	
	/**
	 * cheks whether a tRNA is the cognate tRNA of a Codon
	 * @param codonID
	 * @param tRNAID
	 * @return
	 */
	public boolean isCognate(int codonID, int tRNAID){
            
             
            
		return isCognate[codonID][tRNAID];
	}
	
	/**
	 * cheks whether a tRNA is the near-cognate tRNA of a Codon
	 * @param codonID
	 * @param tRNAID
	 * @return
	 */
	public boolean isNearCognate(int codonID, int tRNAID){
		return isNearCognate[codonID][tRNAID];
	}
	
	/**
	 * returns the total number of tRNAs molecules
	 *
	 */
	public int getTotalAbundance(){
		return totalAbundance;
	}
	
	/**
	 * returns the total number of charged tRNAs
	 * @return
	 */
	public int getChargedAbundance(){
		return this.chargedAbundance;
	}
	
	/**
	 * returns the total number of busy tRNAs
	 * @return
	 */
	public int getBusyAbundance(){
		return this.busyAbundance;
	}
	
	
	/**
	 * returns the total number of uncharged tRNAs
	 * @return
	 */
	public int getUnchargedAbundance(){
		return this.unchargedAbundance;
	}
	
	/**
	 * returns the id of the tRNA of label provided as a parameter
	 * @param label
	 * @return
	 */
	private byte gettRNA(ArrayList<TranslationalRNA> tRNAs, byte label){
		byte result = -1;
		for(byte i=0;i<tRNAs.size();i++){
			if(tRNAs.get(i).getLabel() == label){
				return i;
			}
		}
		return result;
	}
	
	private byte maxtRNALabel(ArrayList<TranslationalRNA> tRNAs){
		byte result = -1;
		for(TranslationalRNA tRNA:tRNAs){
			if(tRNA.getLabel() > result){
				result = tRNA.getLabel();
			}
		}
		return result;
	}
	
	/**
	 * returns the number of tRNA
	 * @return
	 */
	public int getNoOftRNAs(){
		return this.tRNAs.length;
	}
	
	/**
	 * return a specific tRNA
	 * @param id the id of the tRNA
	 * @return
	 */
	public TranslationalRNA gettRNA(int id){
		if(id>=0 && id<tRNAs.length){
			return tRNAs[id];
		} else{
			return null;
		}
	}
	
	/**
	 * array list with the ids of the cognates tRNAs for a specific codon
	 * @param codonID
	 * @return
	 */
	public ArrayList<Integer> getCognates(byte codonID){
		return cognatetRNAs.get(codonID);
	}
	
	/**
	 * array list with the ids of the near cognates tRNAs for a specific codon
	 * @param codonID
	 * @return
	 */
	public ArrayList<Integer> getNearCognates(byte codonID){
		return nearCognatetRNAs.get(codonID);
	}
	
	/**
	 * generates a string table with the tRNA population
	 */
	public String toString(){
		String result=tRNAs.length+"tRNAs\n";
		for(TranslationalRNA tRNA:tRNAs){
			result+=tRNA+"\n";
		}
		
		result+="\nCodons:\n";
		for(int i=0;i<SequenceUtils.codons.codon.length;i++){
			result+=SequenceUtils.codons.codon[i]+": (cognates) ";
			for(int j: cognatetRNAs.get(i)){
				result+=j+", ";
			}
			result+="; (near-cognates) ";
			for(int j: nearCognatetRNAs.get(i)){
				result+=j+", ";
			}
			result+="\n";
		}
		
		return result;
	}
	
}
